package mhtree;

import cz.muni.fi.disa.similarityoperators.cover.AbstractRepresentation.PrecomputedDistances;
import messif.buckets.BucketDispatcher;
import messif.buckets.BucketStorageException;
import messif.objects.LocalAbstractObject;

import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

class BuildTree {

    private final int leafCapacity;
    private final int arity;

    private final Node[] nodes;
    private final BitSet validNodeIndices;

    private final PrecomputedDistances objectDistances;
    private final float[][] nodeDistances;

    private final InsertType insertType;
    private final ObjectToNodeDistance objectToNodeDistance;
    private final BucketDispatcher bucketDispatcher;

    private Node root;

    BuildTree(List<LocalAbstractObject> objects, int leafCapacity, int arity, InsertType insertType, ObjectToNodeDistance objectToNodeDistance, BucketDispatcher bucketDispatcher) throws BucketStorageException {
        this.arity = arity;
        this.leafCapacity = leafCapacity;
        this.insertType = insertType;
        this.objectToNodeDistance = objectToNodeDistance;
        this.bucketDispatcher = bucketDispatcher;

        nodes = new Node[objects.size() / leafCapacity];

        validNodeIndices = new BitSet(nodes.length);
        validNodeIndices.set(0, nodes.length);

        objectDistances = new PrecomputedDistances(objects);
        nodeDistances = new float[nodes.length][nodes.length];

        // Every object is stored in the root
        if (objectDistances.getObjectCount() < leafCapacity) {
            root = new LeafNode(objectDistances, bucketDispatcher.createBucket(), insertType, objectToNodeDistance);
            return;
        }

        createLeafNodes();

        precomputeNodeDistances();

        buildTree();
    }

    public Node getRoot() {
        return root;
    }

    private void buildTree() {
        while (validNodeIndices.cardinality() != 1) {
            BitSet notProcessedNodeIndices = (BitSet) validNodeIndices.clone();

            while (!notProcessedNodeIndices.isEmpty()) {
                if (notProcessedNodeIndices.cardinality() < arity) {
                    Set<Integer> nodeIndices = new HashSet<>(notProcessedNodeIndices.cardinality() - 1);

                    int mainNodeIndex = notProcessedNodeIndices.nextSetBit(0);
                    notProcessedNodeIndices.stream().skip(1).forEach(nodeIndices::add);

                    mergeNodes(mainNodeIndex, nodeIndices);

                    break;
                }

                int furthestNodeIndex = getFurthestIndex(nodeDistances, notProcessedNodeIndices);
                notProcessedNodeIndices.clear(furthestNodeIndex);

                Set<Integer> nnNodeIndices = new HashSet<>(arity - 1);

                for (int i = 0; i < arity - 1; i++) {
                    int index = objectDistances.minDistInArrayExceptIdx(nodeDistances[furthestNodeIndex], notProcessedNodeIndices, furthestNodeIndex);
                    notProcessedNodeIndices.clear(index);
                    nnNodeIndices.add(index);
                }

                mergeNodes(furthestNodeIndex, nnNodeIndices);
            }
        }

        root = nodes[validNodeIndices.nextSetBit(0)];
    }

    private void createLeafNodes() throws BucketStorageException {
        BitSet notProcessedObjectIndices = new BitSet(objectDistances.getObjectCount());
        notProcessedObjectIndices.set(0, objectDistances.getObjectCount());

        for (int nodeIndex = 0; !notProcessedObjectIndices.isEmpty(); nodeIndex++) {
            if (notProcessedObjectIndices.cardinality() < leafCapacity) {
                for (int i = notProcessedObjectIndices.nextSetBit(0); i >= 0; i = notProcessedObjectIndices.nextSetBit(i + 1))
                    addObjectToClosestNode(i);
                return;
            }

            List<LocalAbstractObject> objects = new ArrayList<>(leafCapacity);

            // Select a base object
            int furthestIndex = getFurthestIndex(objectDistances.getDistances(), notProcessedObjectIndices);
            notProcessedObjectIndices.clear(furthestIndex);
            objects.add(objectDistances.getObject(furthestIndex));

            // Select the rest of the objects up to the total of leafCapacity
            objects.addAll(findClosestObjects(furthestIndex, leafCapacity - 1, notProcessedObjectIndices));

            nodes[nodeIndex] = new LeafNode(objectDistances.getSubset(objects), bucketDispatcher.createBucket(), insertType, objectToNodeDistance);
        }
    }

    private void addObjectToClosestNode(int objectIndex) throws BucketStorageException {
        LocalAbstractObject object = objectDistances.getObject(objectIndex);

        Function<Node, Float> getMinHullObjectDistance = node -> node.getHullObjects().stream()
                .map(object::getDistance)
                .reduce(Float.MAX_VALUE, Math::min);

        Map<Node, Float> nodeToObjectDistance = Arrays.stream(nodes)
                .collect(Collectors.toMap(Function.identity(), getMinHullObjectDistance));

        // Add object to the node with minimal distance to this object
        Collections.min(nodeToObjectDistance.entrySet(), Map.Entry.comparingByValue()).getKey().addObject(object);
    }

    private int getFurthestIndex(float[][] distMatrix, BitSet notUsedIndices) {
        float max = Float.MIN_VALUE;
        int maxIndex = notUsedIndices.nextSetBit(0);

        while (true) {
            float[] distances = distMatrix[maxIndex];
            int candidateMaxIdx = this.objectDistances.maxDistInArray(distances, notUsedIndices);
            if (!(distances[candidateMaxIdx] > max)) {
                return maxIndex;
            }

            max = distances[candidateMaxIdx];
            maxIndex = candidateMaxIdx;
        }
    }

    private Set<LocalAbstractObject> findClosestObjects(int baseObjectIndex, int numberOfObjects, BitSet notProcessedIndices) {
        List<Integer> objectIndices = new ArrayList<>(1 + numberOfObjects);

        objectIndices.add(baseObjectIndex);

        return IntStream.range(0, numberOfObjects).mapToObj(i -> {
            HashMap<Integer, Float> indexToDistance = new HashMap<>(objectIndices.size());

            for (int index : objectIndices) {
                int nnIndex = objectDistances.minDistInArray(objectDistances.getDistances(index), notProcessedIndices);

                float distanceSum = objectIndices.stream()
                        .map(objectIndex -> objectDistances.getDistance(objectIndex, nnIndex))
                        .reduce(0f, Float::sum);

                indexToDistance.put(nnIndex, distanceSum);
            }

            int closestPointIndex = Collections.min(indexToDistance.entrySet(), Map.Entry.comparingByValue()).getKey();

            notProcessedIndices.clear(closestPointIndex);
            objectIndices.add(closestPointIndex);

            return objectDistances.getObject(closestPointIndex);
        }).collect(Collectors.toSet());
    }

    private void precomputeNodeDistances() {
        for (int i = 0; i < nodes.length; i++) {
            for (int j = i + 1; j < nodes.length; j++) {
                float minDistance = Float.MAX_VALUE;

                for (LocalAbstractObject firstPoint : nodes[i].getObjects())
                    for (LocalAbstractObject secondPoint : nodes[j].getObjects())
                        minDistance = Math.min(minDistance, objectDistances.getDistance(firstPoint, secondPoint));

                if (minDistance == 0f)
                    throw new RuntimeException("Zero distance between " + nodes[i].toString() + " and " + nodes[j].toString());

                nodeDistances[i][j] = minDistance;
                nodeDistances[j][i] = minDistance;
            }
        }
    }

    private void mergeNodes(int mainNodeIndex, Set<Integer> nodeIndices) {
        if (nodeIndices.size() == 0) return;

        Set<Integer> indices = new HashSet<>(nodeIndices);
        indices.add(mainNodeIndex);

        Set<Node> nodes = indices.stream().map(i -> this.nodes[i]).collect(Collectors.toSet());

        InternalNode parent = Node.createParent(nodes, objectDistances, insertType, objectToNodeDistance);
        nodes.forEach(node -> node.setParent(parent));
        parent.addChildren(nodes);

        this.nodes[mainNodeIndex] = parent;

        for (int i : nodeIndices) {
            validNodeIndices.clear(i);
            this.nodes[i] = null;
        }

        updateNodeDistances(mainNodeIndex, nodeIndices);
    }

    private void updateNodeDistances(int baseNodeIndex, Set<Integer> nodeIndices) {
        if (nodeIndices.size() == 0) return;

        validNodeIndices.stream().forEach(i -> {
            float minDistance = nodeDistances[baseNodeIndex][i];

            for (int index : nodeIndices)
                minDistance = Math.min(minDistance, nodeDistances[index][i]);

            nodeDistances[baseNodeIndex][i] = minDistance;
            nodeDistances[i][baseNodeIndex] = minDistance;
        });
    }
}
