package mhtree;

import messif.algorithms.Algorithm;
import messif.buckets.BucketDispatcher;
import messif.buckets.BucketErrorCode;
import messif.buckets.BucketStorageException;
import messif.buckets.LocalBucket;
import messif.objects.LocalAbstractObject;
import messif.operations.Approximate;
import messif.operations.data.InsertOperation;
import messif.operations.query.ApproxKNNQueryOperation;
import messif.operations.query.KNNQueryOperation;
import messif.statistics.StatisticCounter;
import mhtree.benchmarking.Utils;

import java.io.Serializable;
import java.util.*;

public class MHTree extends Algorithm implements Serializable {

    /**
     * Serialization ID
     */
    private static final long serialVersionUID = 42L;

    private final int leafCapacity;
    private final int arity;

    private final Node root;
    private final BucketDispatcher bucketDispatcher;
    private final InsertType insertType;
    private final ObjectToNodeDistance objectToNodeDistance;

    private final StatisticCounter statVisitedLeaves = StatisticCounter.getStatistics("Node.Leaf.Visited");

    @AlgorithmConstructor(description = "MH-Tree", arguments = {
            "list of objects",
            "number of objects in leaf node",
            "arity",
            "number of threads used in precomputing distances",
            "insert type",
            "distance measure",
            "storage class for buckets",
            "storage class parameters"
    })
    public MHTree(List<LocalAbstractObject> objects, int leafCapacity, int arity, InsertType insertType, ObjectToNodeDistance objectToNodeDistance, Class<? extends LocalBucket> defaultBucketClass, Map<String, Object> bucketClassParams) throws BucketStorageException {
        super("MH-Tree");

        this.leafCapacity = leafCapacity;
        this.arity = arity;
        this.insertType = insertType;
        this.objectToNodeDistance = objectToNodeDistance;

        bucketDispatcher = new BucketDispatcher(Integer.MAX_VALUE, Long.MAX_VALUE, leafCapacity, 0, false, defaultBucketClass, bucketClassParams);

        root = new BuildTree(objects, leafCapacity, arity, insertType, objectToNodeDistance, bucketDispatcher).getRoot();
    }

    public void approxKNN(ApproxKNNQueryOperation operation, double coefficient) {
        LocalAbstractObject queryObject = operation.getQueryObject();

        boolean limitVisitedLeaves = operation.getLocalSearchType() == Approximate.LocalSearchType.DATA_PARTITIONS;

        PriorityQueue<ObjectToNodeDistanceRank> queue = new PriorityQueue<>();
        queue.add(new ObjectToNodeDistanceRank(queryObject, root));

        while (!queue.isEmpty()) {
            Node node = queue.poll().getNode();

            if (node.isLeaf()) {
                if (limitVisitedLeaves && statVisitedLeaves.get() == operation.getLocalSearchParam())
                    break;

                statVisitedLeaves.add();

                for (LocalAbstractObject object : node.getObjects())
                    if (!operation.isAnswerFull() || queryObject.getDistance(object) < operation.getAnswerDistance())
                        operation.addToAnswer(object);
            } else {
                for (Node child : ((InternalNode) node).getChildren())
                    if (!operation.isAnswerFull() || !isPrunable(child, queryObject, operation, coefficient))
                        queue.add(new ObjectToNodeDistanceRank(queryObject, child));
            }
        }

        operation.endOperation();
    }

    private boolean isPrunable(Node child, LocalAbstractObject queryObject, ApproxKNNQueryOperation operation, double coefficient) {
        return operation.getAnswerDistance() * coefficient < child.getNearestDistance(queryObject);
    }

    public void insert(InsertOperation operation) throws BucketStorageException {
        LocalAbstractObject object = operation.getInsertedObject();

        Node node = root;

        while (!node.isLeaf()) {
            node.addObject(object);

            node = ((InternalNode) node).getNearestChild(object);
        }

        node.addObject(object);

        operation.endOperation(BucketErrorCode.OBJECT_INSERTED);
    }

    public void printStatistics() {
        IntSummaryStatistics leafStatistics = bucketDispatcher
                .getAllBuckets()
                .stream()
                .mapToInt(LocalBucket::getObjectCount)
                .summaryStatistics();

        System.out.println("--- STATISTICS ---");

        System.out.println("- MH-Tree -");

        System.out.println("Insert type: " + insertType);
        System.out.println("Height: " + root.getHeight());
        System.out.println("Arity: " + arity);
        System.out.println("Leaf capacity: " + leafCapacity + " objects");

        System.out.println("Number of nodes in each level:");
        for (int level = 1; level <= root.getHeight() + 1; level++) {
            System.out.println("- Level " + level + " -> " + root.getNodesOnLevel(level).size());
        }

        System.out.println("Number of hull objects in each level:");
        for (int level = 1; level <= root.getHeight() + 1; level++) {
            Set<Node> levelNodes = root.getNodesOnLevel(level);

            System.out.println("- Level " + level + " -> " + levelNodes.stream().mapToInt(n -> n.getHullObjects().size()).summaryStatistics());
        }

        System.out.println("\nHistogram of covered objects per level: ");
        System.out.println(Histogram.generate(root));

        System.out.println("- LeafNodes -");

        System.out.println("Count: " + leafStatistics.getCount());
        System.out.println("Minimum number of objects: " + leafStatistics.getMin());
        System.out.println("Average number of objects: " + String.format("%.2f", leafStatistics.getAverage()));
        System.out.println("Maximum number of objects: " + leafStatistics.getMax());
    }

    public long getVisitedLeaves() {
        return statVisitedLeaves.get();
    }

    public void reset() {
        statVisitedLeaves.reset();
    }

    public double measureRecall(ApproxKNNQueryOperation approxKNNQueryOperation) {
        LocalAbstractObject queryObject = approxKNNQueryOperation.getQueryObject();
        int k = approxKNNQueryOperation.getK();

        KNNQueryOperation knnQueryOperation = new KNNQueryOperation(queryObject, k);
        KNN(knnQueryOperation);

        return Utils.measureRecall(approxKNNQueryOperation, knnQueryOperation);
    }

    private void KNN(KNNQueryOperation operation) {
        root.getObjects().forEach(operation::addToAnswer);
        operation.endOperation();
    }

    @Override
    public String toString() {
        return "MHTree{" +
                "leafCapacity=" + leafCapacity +
                ", arity=" + arity +
                ", root=" + root +
                ", insertType" + insertType +
                ", objectToNodeDistance" + objectToNodeDistance +
                '}';
    }
}
